#!/bin/bash
#
# Simple tool providing trash can functionality on Linux console
# in a server environment also.
#
# Part of timecan package.

OVERRIDE_SYSTEM="/etc/default/timecan"	# System-wide config file
OVERRIDE_USER="$HOME/.timecan"		# User specific config file

# Including mandatory, then optional headers.
MSG_TCAN_NOHEADER="Fatal - header file is not found:"
for header in "$OVERRIDE_SYSTEM"
do
    if [ ! -r "$header" ]; then echo -e "$MSG_TCAN_NOHEADER $header" >&2; exit 1; fi
    . "$header"
done
for header in "$OVERRIDE_USER"
do
    if [ -r "$header" ]; then . "$header"; fi
done

# Command call with modifications.
BASENAME=${BASENAME:-$(which basename)}
CP="${CP:-$(which cp)} -v --parents --archive --backup=numbered"
DATE="${DATE:-$(which date)}"
DIRNAME="${DIRNAME:-$(which dirname)}"
FIND="${FIND:-$(which find)}"
RM="${RM:-$(which rm)} -f -v"
RMDIR="${RMDIR:-$(which rmdir)} --ignore-fail-on-non-empty"
SED="${SED:-$(which sed)} -r"
STAT="${STAT:-$(which stat)}"
TOUCH="${TOUCH:-$(which touch)}"
WHOAMI="${WHOAMI:-$(which whoami)}"

# Internal initialisations.
EXIT_ERR=1				# Error signal.
INHERIT=1				# Inherited rules by default.
REMOVE=0				# Not to remove permanently by default.
TIMEFILTER=""				# No timing by default.

# Default settings.
CANDIR=${CANDIR:-"$HOME/.Trash"}                # Trash can base folder
CANDIRMODE=${CANDIRMODE:-"750"}                 # Trash can base folder perms
CANDIROWNER=${CANDIROWNER:-"$USER:$USER"}       # Trash can base folder owner
CONFNAME=${CONFNAME:-".$($BASENAME $0)"}        # Per-folder config file name
LOGFILE=${LOGFILE:-"$HOME/.$($BASENAME $0).log"}    # Logfile pathname
LOGFILEMODE=${LOGFILEMODE:-"640"}                   # Logfile perms
LOGFILEOWNER=${LOGFILEOWNER:-"$USER:$USER"}         # Logfile owner

#
# Functions
#

# Main function: processes an item - a file's, a symlink's or a folder's
# pathname - given as 1st parameter.
function process_item {

    local item="$1"
    local folder
    local result

    # Testing item, on failure giving up.
    # Ignoring not file, not symlink and not folder items.
    if [ ! -f "$item" -a ! -h "$item" -a ! -d "$item" ]; then
	log "(1) $MSG_TCAN_IGNORING $item"; return $EXIT_ERR; fi
    # On file or symlink item we need at least the read permission.
    if [ -f "$item" -o -h "$item" ] && [ ! -r "$item" ]; then
	log "(2) $MSG_TCAN_IGNORING $item"; return $EXIT_ERR; fi
    # On folder item we need at least rx permissions.
    if [ -d "$item" ] && [ ! -r "$item" -o ! -x "$item" ]; then
	log "(3) $MSG_TCAN_IGNORING $item"; return $EXIT_ERR; fi
    # Ignore per-folder config file (always) and veto files (if any).
    for vetofile in "$CONFNAME" $VETOFILES; do
	if [ "$($BASENAME "$item")" = "$vetofile" ]; then
	    log "(4) $MSG_TCAN_IGNORING $item"; return $EXIT_ERR;
	fi
    done

    # Identifying folder containing this item.
    if [ -d "$item" ]; then folder="$item";
    else folder=$($DIRNAME "$item"); fi
    # Parsing per-folder configuration file (if any).
    if [ -r "$folder/$CONFNAME" ]; then parse_config "$folder/$CONFNAME"; fi

    # Determining parameters for this item according to per-folder
    # configuration (if any) and inheritence (if is set).


#    params CONFNAME folder item

    # Recursively processing content of this item, if it's a non-empty folder.
    if [ -d "$item" -a ! -z "$(ls -1A "$item")" ]; then
	params OVERRIDE CANDIR CANDIROWNER CANDIRMODE LOGFILE LOGFILEOWNER LOGFILEMODE REMOVE TIMEFILTER item
	# Processing items within this folder (if any).
	local timestamps=$($STAT --printf="%x\n%y\n%z\n" "$item")
	local filelist=$(ls -1A "$item")
	echo -e "$filelist" | while read filename; do process_item "$item/$filename"; done
	# Processing the folder itself.
	if [ -z "$(ls -A "$item")" ]; then
	    # An emptied source folder will be simply removed.
            message=$($RMDIR "$item" 2>&1); result=$?
            # Only failures will be logged.
            if [ $result -ne 0 ]; then log "($result) $(echo $message)"; fi
	fi
	# Restoring atime and mtime timestamps on source folder.
	if [ -d "$item" ]; then
	    if [ "$($STAT --printf="%x\n" "$item")" != "$(echo -e "$timestamps" | $SED -n '1p')" ]; then
		$TOUCH -a -d "$(echo -e "$timestamps" | $SED -n '1p')" "$item"
	    fi
	    if [ "$($STAT --printf="%y\n" "$item")" != "$(echo -e "$timestamps" | $SED -n '2p')" ]; then
		$TOUCH -m -d "$(echo -e "$timestamps" | $SED -n '2p')" "$item"
	    fi
	fi
	# Synchronising atime and mtime timestamps of trashed folder
	# to the source, if it's content has been changed.
	if [ -d "$CANDIR/$item" -a "$REMOVE" -eq 0 -a "$filelist" != "$(ls -1A "$item" 2>/dev/null)" ]; then
	    if [ "$($STAT --printf="%x\n" "$CANDIR/$item")" != "$(echo -e "$timestamps" | $SED -n '1p')" ]; then
		$TOUCH -a -d "$(echo -e "$timestamps" | $SED -n '1p')" "$CANDIR/$item"
	    fi
	    if [ "$($STAT --printf="%y\n" "$CANDIR/$item")" != "$(echo -e "$timestamps" | $SED -n '2p')" ]; then
		$TOUCH -m -d "$(echo -e "$timestamps" | $SED -n '2p')" "$CANDIR/$item"
	    fi
	fi
	unset filelist timestamps
	# Finished with this folder.

    # This item is an empty folder, a file or a symbolic link.
    else
    	# Applying time filter (if any).
    	if [ ! -z "$TIMEFILTER" ]; then item=$($FIND "$item" $TIMEFILTER)
    	    # If no match, nothing to do with this item.
    	    if [ -z "$item" ]; then return; fi
    	fi
    	# Use trash can method.
    	if [ "$REMOVE" -eq 0 ]; then
    	    # Try to copy, log result.
    	    message=$($CP "$item" $CANDIR 2>&1); result=$?
    	    log "($result) $(echo $message)" # one line
    	    # On success try to remove the source.
    	    if [ $result -eq 0 ]; then
		if [ -d "$item" ];
		    then message=$($RMDIR "$item" 2>&1); result=$?
		    else message=$($RM "$item" 2>&1); result=$?
		fi
        	# Only failures will be logged.
            	if [ $result -ne 0 ]; then log "($result) $(echo $message)"; fi
    	    fi
    	# Use permanent removal method.
    	else
	    if [ -d "$item" ];
		then message=$($RMDIR "$item" 2>&1); result=$?
		else message=$($RM "$item" 2>&1); result=$?
	    fi
    	    log "($result) $(echo $message)" # one line
    	fi
    fi

    return
}

# Transforms a configuration sequence (e.g. textfile as input) into
# shell command line(s) prepared to set some local variables.
# One configuration line from input transformed into one command line to output.
# Valid config line format:
# command tabulator(s) pattern tabulator(s) timefilter
# commands are case insensitive; none, space(s) or comma delimited:
# s for shell, r for regexp, m for move, d for delete, i for inherit, n for not inherit
# Sets following variables: INHERIT, FILTER, PATTERN, REMOVE, TIMEFILTER
# Example input line:
# smi	*	+7
# transformed into output line:
# local INHERIT=1; local FILTER=""; local PATTERN="*"; local REMOVE=0; local TIMEFILTER="+7"
function parse_config {

    return
}

# Helper function: logs with timestamp into a logfile and native to STD_ERR.
# Creates the logfile if absent.
# Consider following variables: LOGFILE, LOGFILEOWNER, LOGFILEMODE
function log {

    # Ignore empty calls.
    if [ -z "$1" ]; then return; fi

    # Local settings.
    local BASENAME=${BASENAME:-$(which basename)}
    local DATE=${DATE:-$(which date)}
    local DATETIME="+%Y-%m-%d %H:%M:%S %Z"
    #local LOGFILE=${LOGFILE:-"/var/log/$($BASENAME $0).log"}
    local LOGFILEMODE="$LOGFILEMODE"
          if [ ! -z "`echo "$LOGFILEMODE" | $SED "s/^[0-7]{3}[0-7]?$//"`" ]; then LOGFILEMODE="640"; fi
    local LOGFILEOWNER="$LOGFILEOWNER"
          if [ ! -z "`echo "$LOGFILEOWNER" | $SED "s/^[0-9a-zA-Z._-]*:[0-9a-zA-Z._-]*$//"`" ]; then LOGFILEOWNER="$USER:$USER"; fi
    local TOUCH=${TOUCH:-$(which touch)}
    local WHOAMI=${WHOAMI:-$(which whoami)}

    # Logfile will be created, if necessary. Any errors are silently ignored.
    if [ ! -w "$LOGFILE" ]; then
	$TOUCH "$LOGFILE" 2>/dev/null
	chown "$LOGFILEOWNER" "$LOGFILE" 2>/dev/null
	chmod "$LOGFILEMODE" "$LOGFILE" 2>/dev/null
    fi
    if [ -f "$LOGFILE" -a -w "$LOGFILE" ]; then
        echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >>$LOGFILE
    #   echo -e "$*" >&2
    else
        echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >&2
    fi

    return
}

# Helper function: logs value of parameters (settings) given.
function params {

    local output=''
    # Try to evaluate parameters into name=value format.
    for param in "$@"
    do
	output="$output $param=$(eval echo \$$param)"
    done
    # Log caller and results (if any).
    if [ ! -z "$output" ]; then log "${FUNCNAME[1]}$output"; fi

    return
}

##############
# Main program
##############

# Getting and parsing switches.
while [ ! -z "$1" -a -z "$(echo "$1" | $SED s/^-[[:alnum:]]*//)" ]
do
    OPTION=$1; shift
    case $OPTION in
        "-C" | "--config" )
            # Reads config file.
            if [ ! -z "$(echo "$1" | $SED "s/ ^-[[:alnum:]]*//")" ]; then
                OVERRIDE=$1; shift
                if [ -r "$OVERRIDE" ]; then
                    . $OVERRIDE
                else
                    # Terminated on failure
                    log "$MSG_TCAN_NOCONFIG $OVERRIDE"
                    exit $EXIT_ERR
                fi
            fi
        ;;
        "-h" | "--help" )
	    # Prints help.
            echo -e "$USAGE"
            exit 0
        ;;
        "-L" | "--log" )
            # Sets logfile.
            if [ -z "$1" -o ! -z "$(echo "$1" | $SED "s/^-[[:alnum:]]*//")" ]; then
                LOGFILE=$1; shift
            fi
        ;;
        "-r" | "--remove" )
            # Sets permanent removal flag.
            REMOVE=1
        ;;
        "-t" | "--time" )
            # Sets time filter.
            if [ ! -z "$(echo "$1" | $SED "s/^-[[:alnum:]]*//")" ]; then
                if [ -z "$(echo "$1" | $SED "s/^\+?[[:digit:]]+[[:space:]]*$//")" ]; then
                    TIMEFILTER="( -mtime $1 -a -ctime $1 )"; shift
                else
                    # Terminated on failure.
                    log "$MSG_TCAN_WRONGTIME $1"
                    exit $EXIT_ERR
                fi
            else
                # Terminated on failure.
                log "$MSG_TCAN_NOTIME"
                exit $EXIT_ERR
            fi
        ;;
        "-T" | "--trashcan" )
            # Sets trashcan folder.
            if [ ! -z "$(echo "$1" | $SED "s/^-[[:alnum:]]*//")" ]; then
                CANDIR=$1; shift
            fi
        ;;
        *)
	    # Unknown option were used.
            log $MSG_TCAN_UNKNOWN $OPTION
            echo -e "$USAGE" >&2
            exit $EXIT_ERR
        ;;
    esac
done
# Emtpy call provides help.
if [ -z "$1" ]; then echo -e "$USAGE" >&2; exit 0; fi

# Logging summarized parameters.
ITEMS="$@"
params OVERRIDE CANDIR CANDIROWNER CANDIRMODE LOGFILE LOGFILEOWNER LOGFILEMODE REMOVE TIMEFILTER ITEMS

# Creating the trashcan if necessary.
if [ $REMOVE -eq 0 -a ! -d "$CANDIR" ]; then
    # Sanitizing folder parameters.
    if [ ! -z "`echo "$CANDIRMODE" | $SED "s/^[0-7]{3}[0-7]?$//"`" ]; then CANDIRMODE="750"; fi
    if [ ! -z "`echo "$CANDIROWNER" | $SED "s/^[0-9a-zA-Z._-]*:[0-9a-zA-Z._-]*$//"`" ]; then CANDIROWNER="$USER:$USER"; fi
    log "$(mkdir --verbose --parents -m "$CANDIRMODE" "$CANDIR" 2>&1)"
    log "$(chown "$CANDIROWNER" "$CANDIR" 2>&1)"
    log "$(chmod "$CANDIRMODE" "$CANDIR" 2>&1)"
fi
# Giving up on failure.
if [ $REMOVE -eq 0 ] && [ ! -d "$CANDIR" -o ! -w "$CANDIR" ]; then
    log "$MSG_TCAN_NOTRASH $CANDIR"
    exit $EXIT_ERR
fi

# Processing items (pathnames, shell patterns, etc. - remaining parameters).
for item in $ITEMS; do process_item "$item"; done

# That's all, folks!
log "Finished."
